@@ -45,13 +45,26 @@
 #ifdef HAVE_NETDB_H
 # include <netdb.h>
 #endif
+/* libbsd if found */
+#ifdef USE_BSD_H
+# include <bsd/string.h>
+#endif /* USE_BSD_H */
+
+/* libstrl if needed */
+#ifdef USE_STRL_H
+# include <strl.h>
+#endif /* USE_STRL_H */
+
+/* opendmarc_strl if needed */
+#ifdef USE_DMARCSTRL_H
+# include <opendmarc_strl.h>
+#endif /* USE_DMARCSTRL_H */
 
 /* libmilter */
 #include <libmilter/mfapi.h>
 
 /* libopendmarc */
 #include <dmarc.h>
-#include <opendmarc_strl.h>
 
 /* opendmarc includes */
 #include "opendmarc.h"
@@ -119,6 +132,10 @@ struct dmarcf_connctx
 	struct sockaddr_storage	cctx_ip;
 	char			cctx_ipstr[BUFRSZ + 1];
 	char			cctx_host[MAXHOSTNAMELEN + 1];
+#if WITH_SPF
+	char			cctx_helo[MAXHOSTNAMELEN + 1];
+	char			cctx_rawmfrom[MAXHOSTNAMELEN + 1];
+#endif
 };
 typedef struct dmarcf_connctx * DMARCF_CONNCTX;
 
@@ -134,6 +151,11 @@ struct dmarcf_config
 	_Bool			conf_addswhdr;
 	_Bool			conf_authservidwithjobid;
 	_Bool			conf_recordall;
+#if WITH_SPF
+	_Bool			conf_spfignoreresults;
+	_Bool			conf_spfselfvalidate;
+#endif /* WITH_SPF */
+	_Bool			conf_ignoreauthclients;
 	unsigned int		conf_refcnt;
 	unsigned int		conf_dnstimeout;
 	struct config *		conf_data;
@@ -193,6 +215,9 @@ struct lookup log_facilities[] =
 sfsistat mlfi_abort __P((SMFICTX *));
 sfsistat mlfi_close __P((SMFICTX *));
 sfsistat mlfi_connect __P((SMFICTX *, char *, _SOCK_ADDR *));
+#if WITH_SPF
+sfsistat mlfi_helo __P((SMFICTX *, char *));
+#endif
 sfsistat mlfi_envfrom __P((SMFICTX *, char **));
 sfsistat mlfi_eom __P((SMFICTX *));
 sfsistat mlfi_header __P((SMFICTX *, char *, char *));
@@ -743,7 +768,7 @@ dmarcf_mkarray(char *str, char ***array)
 	int ns;
 	char *p;
 	char *ctx;
-	char **out;
+	char **out = NULL;
 
 	for (p = strtok_r(str, ",", &ctx);
 	     p != NULL;
@@ -1223,6 +1248,16 @@ dmarcf_config_load(struct config *data, struct dmarcf_config *conf,
 		                  &conf->conf_enablecores,
 		                  sizeof conf->conf_enablecores);
 
+#if WITH_SPF
+		(void) config_get(data, "SPFIgnoreResults",
+		                  &conf->conf_spfignoreresults,
+		                  sizeof conf->conf_spfignoreresults);
+
+		(void) config_get(data, "SPFSelfValidate",
+		                  &conf->conf_spfselfvalidate,
+		                  sizeof conf->conf_spfselfvalidate);
+#endif /* WITH_SPF */
+
 		(void) config_get(data, "RejectFailures",
 		                  &conf->conf_rejectfail,
 		                  sizeof conf->conf_rejectfail);
@@ -1231,19 +1266,19 @@ dmarcf_config_load(struct config *data, struct dmarcf_config *conf,
 		                  &conf->conf_reqhdrs,
 		                  sizeof conf->conf_reqhdrs);
 
-		(void) config_get(data, "ForensicReports",
+		(void) config_get(data, "FailureReports",
 		                  &conf->conf_afrf,
 		                  sizeof conf->conf_afrf);
 
-		(void) config_get(data, "ForensicReportsOnNone",
+		(void) config_get(data, "FailureReportsOnNone",
 		                  &conf->conf_afrfnone,
 		                  sizeof conf->conf_afrfnone);
 
-		(void) config_get(data, "ForensicReportsSentBy",
+		(void) config_get(data, "FailureReportsSentBy",
 		                  &conf->conf_afrfas,
 		                  sizeof conf->conf_afrfas);
 
-		(void) config_get(data, "ForensicReportsBcc",
+		(void) config_get(data, "FailureReportsBcc",
 		                  &conf->conf_afrfbcc,
 		                  sizeof conf->conf_afrfbcc);
 
@@ -1251,6 +1286,10 @@ dmarcf_config_load(struct config *data, struct dmarcf_config *conf,
 		                  &conf->conf_recordall,
 		                  sizeof conf->conf_recordall);
 
+		(void) config_get(data, "IgnoreAuthenticatedClients",
+		                  &conf->conf_ignoreauthclients,
+		                  sizeof conf->conf_ignoreauthclients);
+
 		(void) config_get(data, "TemporaryDirectory",
 		                  &conf->conf_tmpdir,
 		                  sizeof conf->conf_tmpdir);
@@ -1294,10 +1333,8 @@ dmarcf_config_load(struct config *data, struct dmarcf_config *conf,
 		}
 	}
 
-	if (conf->conf_authservid == NULL)
-		conf->conf_authservid = strdup(myhostname);
-
-	if (conf->conf_trustedauthservids == NULL)
+	if (conf->conf_trustedauthservids == NULL &&
+	    conf->conf_authservid != NULL)
 	{
 		dmarcf_mkarray(conf->conf_authservid,
 		               &conf->conf_trustedauthservids);
@@ -1547,7 +1584,7 @@ mlfi_negotiate(SMFICTX *ctx,
 	unsigned long *pf0, unsigned long *pf1,
 	unsigned long *pf2, unsigned long *pf3)
 {
-	unsigned long reqactions = SMFIF_ADDHDRS;
+	unsigned long reqactions = SMFIF_ADDHDRS|SMFIF_QUARANTINE;
 	unsigned long wantactions = 0;
 	unsigned long protosteps = (SMFIP_NOHELO |
 	                            SMFIP_NOUNKNOWN |
@@ -1746,6 +1783,43 @@ mlfi_connect(SMFICTX *ctx, char *host, _SOCK_ADDR *ip)
 	return SMFIS_CONTINUE;
 }
 
+#if WITH_SPF
+/*
+**  MLFI_HELO -- handler for HELO/EHLO command; only used for spf checks if configured.
+**
+**  Parameters:
+**  	ctx -- milter context
+**  	helo_domain -- possible helo domain
+**
+**  Return value:
+**  	An SMFIS_* constant.
+*/
+
+sfsistat
+mlfi_helo(SMFICTX *ctx, char *helo_domain)
+{
+	DMARCF_CONNCTX cc;
+	struct dmarcf_config *conf;
+
+	assert(ctx != NULL);
+
+	cc = (DMARCF_CONNCTX) dmarcf_getpriv(ctx);
+	if (cc != NULL)
+	{
+		conf = cc->cctx_config;
+		if (!conf->conf_spfselfvalidate)
+			return SMFIS_CONTINUE;
+
+		if (helo_domain != NULL)
+		{
+			strncpy(cc->cctx_helo, helo_domain,
+			        sizeof cc->cctx_helo - 1);
+		}
+	}
+	return SMFIS_CONTINUE;
+}
+#endif /* WITH_SPF */
+
 /*
 **  MLFI_ENVFROM -- handler for MAIL FROM command; used to reset for a message
 **
@@ -1773,6 +1847,10 @@ mlfi_envfrom(SMFICTX *ctx, char **envfrom)
 	if (cc->cctx_msg != NULL)
 		dmarcf_cleanup(ctx);
 
+	if (conf->conf_ignoreauthclients &&
+	    dmarcf_getsymval(ctx, "auth_authen") != NULL)
+		return SMFIS_ACCEPT;
+
 	dfc = (DMARCF_MSGCTX) malloc(sizeof(struct dmarcf_msgctx));
 	if (dfc == NULL)
 	{
@@ -1807,6 +1885,10 @@ mlfi_envfrom(SMFICTX *ctx, char **envfrom)
 		unsigned char *p;
 		unsigned char *q;
 
+#if WITH_SPF
+		strncpy(cc->cctx_rawmfrom, envfrom[0],
+			sizeof cc->cctx_rawmfrom - 1);
+#endif
 		strncpy(dfc->mctx_envfrom, envfrom[0],
 		        sizeof dfc->mctx_envfrom - 1);
 
@@ -1993,12 +2075,15 @@ mlfi_eom(SMFICTX *ctx)
 	if (authservid == NULL)
 		authservid = hostname;
 
+	/* ensure there was a From field */
+	from = dmarcf_findheader(dfc, "From", 0);
+
 	/* if requested, verify RFC5322-required headers (RFC5322 3.6) */
 	if (conf->conf_reqhdrs)
 	{
 		unsigned char* reqhdrs_error = NULL; /* no error */
 
-		if (dmarcf_findheader(dfc, "From", 0) == NULL ||
+		if (from == NULL ||
 		    dmarcf_findheader(dfc, "From", 1) != NULL)
 			reqhdrs_error = "not exactly one From field";
 
@@ -2043,8 +2128,20 @@ mlfi_eom(SMFICTX *ctx)
 		}
 	}
 
+	/* if there was no From:, there's nothing to process past here */
+	if (from == NULL)
+	{
+		if (conf->conf_dolog)
+		{
+			syslog(LOG_INFO,
+			       "%s: RFC5322 requirement error: missing From field; accepting",
+			       dfc->mctx_jobid);
+		}
+
+		return SMFIS_ACCEPT;
+	}
+
 	/* extract From: domain */
-	from = dmarcf_findheader(dfc, "From", 0);
 	memset(addrbuf, '\0', sizeof addrbuf);
 	strncpy(addrbuf, from->hdr_value, sizeof addrbuf - 1);
 	status = dmarcf_mail_parse(addrbuf, &user, &domain);
@@ -2119,8 +2216,10 @@ mlfi_eom(SMFICTX *ctx)
 			continue;
 
 		/* skip it if it's not one of ours */
-		if (!dmarcf_match(ar.ares_host, conf->conf_trustedauthservids,
-		                  FALSE))
+		if (strcasecmp(ar.ares_host, authservid) != 0 &&
+		    (conf->conf_trustedauthservids == NULL ||
+		     !dmarcf_match(ar.ares_host, conf->conf_trustedauthservids,
+		                   FALSE)))
 		{
 			unsigned char *slash;
 
@@ -2152,9 +2251,11 @@ mlfi_eom(SMFICTX *ctx)
 			}
 
 			*slash = '\0';
-			if (!dmarcf_match(ar.ares_host,
-			                  conf->conf_trustedauthservids,
-			                  FALSE) ||
+			if ((strcasecmp(ar.ares_host, authservid) != 0 &&
+			     (conf->conf_trustedauthservids == NULL ||
+			      !dmarcf_match(ar.ares_host,
+			                    conf->conf_trustedauthservids,
+			                    FALSE))) ||
 			    strcmp(slash + 1, dfc->mctx_jobid) != 0)
 			{
 				*slash = '/';
@@ -2169,12 +2270,18 @@ mlfi_eom(SMFICTX *ctx)
 
 				continue;
 			}
+
+			*slash = '/';
 		}
 
 		/* walk through what was found */
 		for (c = 0; c < ar.ares_count; c++)
 		{
-			if (ar.ares_result[c].result_method == ARES_METHOD_SPF)
+			if (ar.ares_result[c].result_method == ARES_METHOD_SPF
+#if WITH_SPF
+			    && !conf->conf_spfignoreresults
+#endif
+			)
 			{
 				int spfmode;
 
@@ -2327,7 +2434,11 @@ mlfi_eom(SMFICTX *ctx)
 		}
 	}
 
-	if (!wspf)
+	if (!wspf
+#if WITH_SPF
+	    && !conf->conf_spfignoreresults
+#endif
+	)
 	{
 		for (hdr = dfc->mctx_hqhead;
 		     hdr != NULL && !wspf;
@@ -2359,13 +2470,111 @@ mlfi_eom(SMFICTX *ctx)
 		}
 	}
 
-	if (!wspf)
-		dmarcf_dstring_printf(dfc->mctx_histbuf, "spf -1\n");
-
 	/*
 	**  Interact with libopendmarc.
 	*/
 
+	if (!wspf)
+	{
+#if WITH_SPF
+		if (conf->conf_spfselfvalidate)
+		{
+			int spf_result;
+			char human[512];
+			int used_mfrom;
+			char *use_domain;
+			int spf_mode;
+			char *pass_fail;
+
+# if HAVE_SPF2_H
+			spf_result = opendmarc_spf2_test(
+# else /* HAVE_SPF2_H */
+			spf_result = opendmarc_spf_test(
+# endif /* HAVE_SPF2_H */
+				cc->cctx_ipstr,
+				cc->cctx_rawmfrom,
+				cc->cctx_helo,
+				NULL,
+				FALSE,
+				human,
+				sizeof human,
+				&used_mfrom);
+			if (used_mfrom == TRUE)
+			{
+				use_domain = dfc->mctx_envfrom;
+				spf_mode   = DMARC_POLICY_SPF_ORIGIN_MAILFROM;
+			}
+			else
+			{
+				use_domain = cc->cctx_helo;
+				spf_mode   = DMARC_POLICY_SPF_ORIGIN_HELO;
+			}
+			ostatus = opendmarc_policy_store_spf(cc->cctx_dmarc, 
+				                                     use_domain,
+				                                     spf_result,
+				                                     spf_mode,
+				                                     human);
+			switch (spf_mode)
+			{
+			    case DMARC_POLICY_SPF_OUTCOME_PASS:
+				pass_fail = "pass";
+				dfc->mctx_spfresult = ARES_RESULT_PASS;
+				break;
+
+			    case DMARC_POLICY_SPF_OUTCOME_NONE:
+				pass_fail = "none";
+				dfc->mctx_spfresult = ARES_RESULT_NONE;
+				break;
+
+			    case DMARC_POLICY_SPF_OUTCOME_TMPFAIL:
+				pass_fail = "tempfail";
+				dfc->mctx_spfresult = ARES_RESULT_TEMPERROR;
+				break;
+
+			    case DMARC_POLICY_SPF_OUTCOME_FAIL:
+				dfc->mctx_spfresult = ARES_RESULT_FAIL;
+				pass_fail = "fail";
+				break;
+
+			    default:
+				dfc->mctx_spfresult = ARES_RESULT_PERMERROR;
+				pass_fail = "permerror";
+				break;
+			}
+
+			if (spf_mode == DMARC_POLICY_SPF_ORIGIN_HELO)
+			{
+				snprintf(header, sizeof header,
+					 "%s; spf=%s smtp.helo=%s",
+					 authservid, pass_fail, use_domain);
+			}
+			else
+			{
+				snprintf(header, sizeof header,
+					 "%s; spf=%s smtp.mailfrom=%s",
+					 authservid, pass_fail, use_domain);
+			}
+
+			if (dmarcf_insheader(ctx, 1, AUTHRESULTSHDR,
+					     header) == MI_FAILURE)
+			{
+				if (conf->conf_dolog)
+				{
+					syslog(LOG_ERR,
+					       "%s: %s header add failed",
+					       dfc->mctx_jobid,
+					       AUTHRESULTSHDR);
+				}
+			}
+
+			dmarcf_dstring_printf(dfc->mctx_histbuf, "spf %d\n",
+			                      dfc->mctx_spfresult);
+		}
+		else
+#endif /* WITH_SPF */
+			dmarcf_dstring_printf(dfc->mctx_histbuf, "spf -1\n");
+	}
+
 	ostatus = opendmarc_policy_query_dmarc(cc->cctx_dmarc,
 	                                       dfc->mctx_fromdomain);
 	if (ostatus == DMARC_PARSE_ERROR_NULL_CTX ||
@@ -2457,7 +2666,7 @@ mlfi_eom(SMFICTX *ctx)
 	dmarcf_dstring_printf(dfc->mctx_histbuf, "align_spf %d\n", align_spf);
 
 	/*
-	**  Generate a forensic report.
+	**  Generate a failure report.
 	*/
 
 	ruv = opendmarc_policy_fetch_ruf(cc->cctx_dmarc, NULL, 0, TRUE);
@@ -2478,7 +2687,7 @@ mlfi_eom(SMFICTX *ctx)
 				if (conf->conf_dolog)
 				{
 					syslog(LOG_ERR,
-					       "%s: failed to create buffer for forensic report",
+					       "%s: unable to create buffer for failure report",
 					       dfc->mctx_jobid);
 				}
 
@@ -2493,7 +2702,7 @@ mlfi_eom(SMFICTX *ctx)
 		if (conf->conf_afrfas != NULL)
 		{
 			dmarcf_dstring_printf(dfc->mctx_afrf,
-			                      "From: %s",
+			                      "From: %s\n",
 			                      conf->conf_afrfas);
 		}
 		else
@@ -2580,7 +2789,7 @@ mlfi_eom(SMFICTX *ctx)
 
 			dmarcf_dstring_printf(dfc->mctx_afrf,
 			                      "Content-Type: multipart/report;"
-			                      "\n\treport-type=feedbackreport;"
+			                      "\n\treport-type=feedback-report;"
 			                      "\n\tboundary=\"%s:%s\"\n",
 			                      hostname, dfc->mctx_jobid);
 
@@ -2671,8 +2880,27 @@ mlfi_eom(SMFICTX *ctx)
 				status = pclose(out);
 				if (status != 0 && conf->conf_dolog)
 				{
-					syslog(LOG_ERR, "%s: pclose() returned status %d",
-					       dfc->mctx_jobid, status);
+					int val;
+					const char *how;
+
+					if (WIFEXITED(status))
+					{
+						how = "exited with status";
+						val = WEXITSTATUS(status);
+					}
+					else if (WIFSIGNALED(status))
+					{
+						how = "killed with signal";
+						val = WTERMSIG(status);
+					}
+					else
+					{
+						how = "returned status";
+						val = status;
+					}
+
+					syslog(LOG_ERR, "%s: pclose() %s %d",
+					       dfc->mctx_jobid, how, val);
 				}
 			}
 		}
@@ -2964,7 +3192,11 @@ struct smfiDesc smfilter =
 	SMFI_VERSION,	/* version code -- do not change */
 	0,		/* flags; updated in main() */
 	mlfi_connect,	/* connection info filter */
+#if WITH_SPF
+	mlfi_helo,	/* SMTP HELO command filter */
+#else
 	NULL,		/* SMTP HELO command filter */
+#endif
 	mlfi_envfrom,	/* envelope sender filter */
 	NULL,		/* envelope recipient filter */
 	mlfi_header,	/* header filter */
@@ -4281,14 +4513,26 @@ main(int argc, char **argv)
 		       VERSION, argstr);
 
 		memset(argstr, '\0', sizeof argstr);
+		strlcpy(argstr, "(none)", sizeof argstr);
 		n = sizeof argstr;
-		for (c = 0; curconf->conf_trustedauthservids[c] != NULL; c++)
+		for (c = 0;
+		     curconf->conf_trustedauthservids != NULL &&
+		     curconf->conf_trustedauthservids[c] != NULL;
+		     c++)
 		{
-			if (c != 0)
+			if (c == 0)
+			{
+				strlcpy(argstr, 
+				        curconf->conf_trustedauthservids[c],
+				        n);
+			}
+			else
+			{
 				strlcat(argstr, ", ", n);
-			strlcat(argstr,
-			        curconf->conf_trustedauthservids[c],
-			        n);
+				strlcat(argstr,
+				        curconf->conf_trustedauthservids[c],
+				        n);
+			}
 		}
 
 		syslog(LOG_INFO, "trusted authentication services: %s",
